@@ -14,7 +14,6 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-
 import sys
 from PyQt5.QtCore import QCoreApplication, QLocale, QTranslator
 from PyQt5.QtWidgets import QApplication
@@ -25,7 +24,7 @@ import gpxviewer.rc_gpxviewer
 def main():
   app = QApplication(sys.argv)
   QCoreApplication.setApplicationName('GPX Viewer')
-  QCoreApplication.setApplicationVersion('2.1')
+  QCoreApplication.setApplicationVersion('2.2')
 
   if app.palette().window().color().lightness() < app.palette().windowText().color().lightness():
     TheConfig['MainWindow']['ColorTheme'] = 'dark_theme'
@@ -42,7 +41,7 @@ def main():
   gpxMainWindow = GpxMainWindow()
   gpxMainWindow.show()
   if len(sys.argv) > 1:
-    if sys.argv[1].lower().endswith('.gpx'):
+    if sys.argv[1].lower().endswith('.gpx') or sys.argv[1].lower().endswith('.kml'):
       gpxMainWindow.openGPXFiles(sys.argv[1:])
     else:
       gpxMainWindow.openGPXProject(sys.argv[1])
