@@ -78,6 +78,12 @@ class GpxDocument(QObject):
     self.doc['ChangedNames'] = list(self.wptmodel.changedNames.keys())
     self.doc['PointNames'] = list(self.wptmodel.changedNames.values())
 
+    self.doc['ChangedAltitudes'] = list(self.wptmodel.changedAltitudes.keys())
+    self.doc['PointAltitudes'] = list(self.wptmodel.changedAltitudes.values())
+
+    self.doc['ChangedTrackAltitudes'] = list(self.trkmodel.changedAltitudes.keys())
+    self.doc['TrackAltitudes'] = self.trkmodel.getChangedAltitudes()
+
     cfg = GpxConfigParser()
 
     if self.projectType == TYPE_GPXV:
@@ -246,6 +252,11 @@ class GpxDocument(QObject):
       if 'ChangedNames' in self.doc and 'PointNames' in self.doc:
         for i, n in zip(self.doc['ChangedNames'], self.doc['PointNames']):
           self.wptmodel.setData(self.wptmodel.index(i, gpx.NAME), n, Qt.EditRole)
+      if 'ChangedAltitudes' in self.doc and 'PointAltitudes' in self.doc:
+        self.wptmodel.setAltitudes(self.doc['ChangedAltitudes'], self.doc['PointAltitudes'])
+
+      if 'ChangedTrackAltitudes' in self.doc and 'TrackAltitudes' in self.doc:
+        self.trkmodel.setTracksAltitudes(self.doc['ChangedTrackAltitudes'], self.doc['TrackAltitudes'])
     else:  # GPXMAGICK not in cfg
       raise gpx.GpxWarning(QFileInfo(filename).absoluteFilePath() + QCoreApplication.translate('GpxDocument', ' is not a valid GPX Viewer project.'))
 
