@@ -17,8 +17,9 @@
 import sys
 import re
 import webbrowser
-from PyQt5.QtCore import Qt, QCoreApplication, QEvent, QFileInfo, QFileSelector, QT_VERSION_STR, PYQT_VERSION_STR, pyqtSlot
+from PyQt5.QtCore import Qt, QCoreApplication, QEvent, QFileInfo, QFileSelector, QUrl, QT_VERSION, QT_VERSION_STR, PYQT_VERSION_STR, pyqtSlot
 from PyQt5.QtGui import QCursor, QIcon, QKeyEvent, QKeySequence
+from PyQt5.QtNetwork import QNetworkAccessManager, QNetworkReply, QNetworkRequest
 from PyQt5.QtWidgets import (QAction, QApplication, QDialog, QDialogButtonBox, QFileDialog, QHBoxLayout, QLabel, QLineEdit,
                              QMainWindow, QMenu, QMessageBox, QSizePolicy, QSpacerItem, QSpinBox, QVBoxLayout, QWidget)
 import gpxviewer.gpxmodel as gpx
@@ -65,6 +66,11 @@ class GpxMainWindow(QMainWindow):
     self.ui.wptView.installEventFilter(self)
     self.ui.trkView.installEventFilter(self)
 
+    self.networkManager = QNetworkAccessManager(self)
+    if QT_VERSION >= 0x050f00:  # function introduced in 5.15.0
+      self.networkManager.setTransferTimeout(10000)
+    self.networkManager.finished[QNetworkReply].connect(self.replyReady)
+
     self.ui.actionDetailedView.setChecked(TheConfig['MainWindow'].getboolean('DetailedView'))
     self.ui.actionShowSkipped.setChecked(TheConfig['MainWindow'].getboolean('ShowSkipped'))
     self.ui.actionShowMarked.setChecked(TheConfig['MainWindow'].getboolean('ShowMarked'))
@@ -75,6 +81,7 @@ class GpxMainWindow(QMainWindow):
     self.ui.tabWidget.currentChanged[int].connect(self.onTabChanged)
     TheDocument.gpxparser.warningSent[str, str].connect(self.showWarning)
     TheDocument.wptmodel.wptDataChanged.connect(self.setProjectChanged)
+    TheDocument.trkmodel.trkDataChanged.connect(self.setProjectChanged)
     TheDocument.fileNotFound[str].connect(self.openedFileNotFound)
     TheDocument.askToExtractFiles[list].connect(self.askedExtractFiles)
 
@@ -97,8 +104,8 @@ class GpxMainWindow(QMainWindow):
     self.themeSelector.setExtraSelectors([TheConfig['MainWindow']['ColorTheme']])
 
     self.setWindowIcon(QIcon(':/icons/gpxviewer.svg'))
-    self.ui.actionLoadGPXfile.setIcon(QIcon(self.themeSelector.select(':/icons/xml-element-new.svg')))
-    self.ui.actionSaveGPXfileAs.setIcon(QIcon(self.themeSelector.select(':/icons/document-save-as-template.svg')))
+    self.ui.actionLoadFile.setIcon(QIcon(self.themeSelector.select(':/icons/xml-element-new.svg')))
+    self.ui.actionSaveFileAs.setIcon(QIcon(self.themeSelector.select(':/icons/document-save-as-template.svg')))
     self.ui.actionNew.setIcon(QIcon(self.themeSelector.select(':/icons/document-new.svg')))
     self.ui.actionOpen.setIcon(QIcon(self.themeSelector.select(':/icons/document-open.svg')))
     self.ui.actionSave.setIcon(QIcon(self.themeSelector.select(':/icons/document-save.svg')))
@@ -181,6 +188,16 @@ class GpxMainWindow(QMainWindow):
     self.actStyle.triggered.connect(self.pointStyle)
     self.addAction(self.actStyle)
 
+    self.actGetAltitude = QAction(QIcon(self.themeSelector.select(':/icons/download.svg')), self.tr('Get altitudes from SRTM'), self)
+    self.actGetAltitude.setStatusTip(self.tr('Get altitudes from online SRTM data'))
+    self.actGetAltitude.triggered.connect(self.requestAltitudes)
+    self.addAction(self.actGetAltitude)
+
+    self.actResetAltitude = QAction(QIcon(self.themeSelector.select(':/icons/edit-clear.svg')), self.tr('Reset altitude'), self)
+    self.actResetAltitude.setStatusTip(self.tr('Reset the altitudes of these waypoints'))
+    self.actResetAltitude.triggered.connect(self.resetAltitudes)
+    self.addAction(self.actResetAltitude)
+
     self.actSkipTracks = QAction(QIcon(self.themeSelector.select(':/icons/waypoint-skip.svg')), self.tr('Skip tracks'), self)
     self.actSkipTracks.setStatusTip(self.tr('Skip these tracks when plotting profiles or calculating statistics'))
     self.actSkipTracks.setShortcut(Qt.Key_Delete)
@@ -188,6 +205,16 @@ class GpxMainWindow(QMainWindow):
     self.actSkipTracks.triggered.connect(self.skipTracks)
     self.addAction(self.actSkipTracks)
 
+    self.actGetTrackAltitude = QAction(QIcon(self.themeSelector.select(':/icons/download.svg')), self.tr('Get altitudes from SRTM'), self)
+    self.actGetTrackAltitude.setStatusTip(self.tr('Get altitudes from online SRTM data'))
+    self.actGetTrackAltitude.triggered.connect(self.requestTrackAltitudes)
+    self.addAction(self.actGetTrackAltitude)
+
+    self.actResetTrackAltitude = QAction(QIcon(self.themeSelector.select(':/icons/edit-clear.svg')), self.tr('Reset altitude'), self)
+    self.actResetTrackAltitude.setStatusTip(self.tr('Reset the altitudes of these tracks'))
+    self.actResetTrackAltitude.triggered.connect(self.resetTrackAltitudes)
+    self.addAction(self.actResetTrackAltitude)
+
   @pyqtSlot()
   def onAboutQt(self):
     QApplication.aboutQt()
@@ -246,6 +273,8 @@ class GpxMainWindow(QMainWindow):
       menu.addSeparator()
       menu.addAction(self.actStyle)
       menu.addSeparator()
+      menu.addAction(self.actGetAltitude)
+      menu.addAction(self.actResetAltitude)
 
       actShowGoogleMap = QAction(QIcon(':/icons/googlemaps.png'), self.tr('Google Maps'), self)
       actShowGoogleMap.setStatusTip(self.tr('Show this waypoint on the website maps.google.com'))
@@ -285,6 +314,9 @@ class GpxMainWindow(QMainWindow):
 
       menu = QMenu(self)
       menu.addAction(self.actSkipTracks)
+      menu.addSeparator()
+      menu.addAction(self.actGetTrackAltitude)
+      menu.addAction(self.actResetTrackAltitude)
       menu.popup(QCursor.pos())
       event.accept()
 
@@ -327,7 +359,7 @@ class GpxMainWindow(QMainWindow):
       QMessageBox.warning(self, self.tr('File open error'), self.tr('Can\'t transform recieved data to local file names.'))
       return
 
-    if filenames[0].lower().endswith('.gpx'):
+    if filenames[0].lower().endswith('.gpx') or filenames[0].lower().endswith('.kml'):
       self.openGPXFiles(filenames)
     elif self.closeChangedProject(self.tr('Open GPX Viewer project')):
       self.openGPXProject(filenames[0])
@@ -387,35 +419,35 @@ class GpxMainWindow(QMainWindow):
   @pyqtSlot()
   def skipPoints(self):
     included = self.ui.wptView.currentIndex().data(gpx.IncludeRole)
-    TheDocument.wptmodel.setIncludeStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME], not included)
+    TheDocument.wptmodel.setIncludeStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()], not included)
     self.filterModel.invalidateFilter()
 
   @pyqtSlot()
   def markerPoints(self):
     marked = self.ui.wptView.currentIndex().data(gpx.MarkerRole)
-    TheDocument.wptmodel.setMarkerStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME], not marked)
+    TheDocument.wptmodel.setMarkerStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()], not marked)
     self.filterModel.invalidateFilter()
 
   @pyqtSlot()
   def captionPoints(self):
     captioned = self.ui.wptView.currentIndex().data(gpx.CaptionRole)
-    TheDocument.wptmodel.setCaptionStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME], not captioned)
+    TheDocument.wptmodel.setCaptionStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()], not captioned)
     self.filterModel.invalidateFilter()
 
   @pyqtSlot()
   def splitLines(self):
     splited = self.ui.wptView.currentIndex().data(gpx.SplitLineRole)
-    TheDocument.wptmodel.setSplitLines([i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME], not splited)
+    TheDocument.wptmodel.setSplitLines([i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()], not splited)
     self.filterModel.invalidateFilter()
 
   @pyqtSlot()
   def neglectDistance(self):
     neglected = self.ui.wptView.currentIndex().data(gpx.NeglectRole)
-    TheDocument.wptmodel.setNeglectStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME], not neglected)
+    TheDocument.wptmodel.setNeglectStates([i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()], not neglected)
 
   @pyqtSlot()
   def resetPoints(self):
-    indexes = [i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME]
+    indexes = [i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()]
     TheDocument.wptmodel.setMarkerStates(indexes, False)
     TheDocument.wptmodel.setCaptionStates(indexes, False)
     TheDocument.wptmodel.setSplitLines(indexes, False)
@@ -424,7 +456,7 @@ class GpxMainWindow(QMainWindow):
 
   @pyqtSlot()
   def renamePoints(self):
-    indexes = [i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME]
+    indexes = [i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()]
     indexes.sort()
     first_name = TheDocument.wptmodel.index(indexes[0], gpx.NAME).data()
 
@@ -462,48 +494,125 @@ class GpxMainWindow(QMainWindow):
 
   @pyqtSlot()
   def resetPointNames(self):
-    TheDocument.wptmodel.resetNames([i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME])
+    TheDocument.wptmodel.resetNames([i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()])
+
+  @pyqtSlot()
+  def requestAltitudes(self):
+    allIndexes = [i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()]
+    maxCount = 20  # maximum number of coordinates requested at once
+    cumNumber = 0
+    self.requestedIndexes = {}
+    self.showNetworkError = True
+
+    while cumNumber < len(allIndexes):
+      num = min(len(allIndexes) - cumNumber, maxCount)
+      indexes = allIndexes[cumNumber:cumNumber + num]
+      cumNumber += num
+      lats = [str(TheDocument.wptmodel.index(i, gpx.LAT).data(gpx.ValueRole)) for i in indexes]
+      lons = [str(TheDocument.wptmodel.index(i, gpx.LON).data(gpx.ValueRole)) for i in indexes]
+      request = 'http://api.geonames.org/srtm1?lats=' + ','.join(lats) + '&lngs=' + ','.join(lons) + '&username=gpxviewer'
+      self.requestedIndexes[request] = indexes
+      self.networkManager.get(QNetworkRequest(QUrl(request)))
+      self.ui.statusBar.showMessage(self.tr('Getting altitudes from SRTM data...'))
+
+  @pyqtSlot(QNetworkReply)
+  def replyReady(self, reply):
+    if reply.error() != QNetworkReply.NoError and self.showNetworkError:
+      self.showNetworkError = False
+      QMessageBox.warning(self, self.tr('Network error'),
+                          self.tr('Error requesting SRTM data from http://geonames.org.') + '\n\n' + reply.errorString())
+
+    else:
+      request = reply.request().url().toString()
+      try:
+        indexes = self.requestedIndexes.pop(request)
+        alts = []
+        while not reply.atEnd():
+          line = reply.readLine()
+          alts += [int(line)]
+
+        if type(indexes) == list:  # waypoints
+          TheDocument.wptmodel.setAltitudes(indexes, alts)
+        elif type(indexes) == tuple:  # tracks
+          TheDocument.trkmodel.setPointsAltitudes(indexes[0], indexes[1], indexes[2], alts)
+
+      except ValueError:
+        self.showNetworkError = False
+        QMessageBox.warning(self, self.tr('Network error'),
+                            self.tr('Error requesting SRTM data from http://geonames.org.') + '\n\n' + str(line, 'utf-8'))
+
+      if len(self.requestedIndexes) == 0:
+        self.ui.statusBar.showMessage(self.tr('Getting altitudes from SRTM data finished.'), 2000)
+
+  @pyqtSlot()
+  def resetAltitudes(self):
+    TheDocument.wptmodel.resetAltitudes([i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()])
+    self.ui.statusBar.showMessage(self.tr('Altitudes reset to original values.'), 2000)
+
+  @pyqtSlot()
+  def requestTrackAltitudes(self):
+    track = self.ui.trkView.currentIndex().row()
+    maxCount = 20  # maximum number of coordinates requested at once
+    self.requestedIndexes = {}
+
+    for s, segment in enumerate(TheDocument.trkmodel.tracks[track]['SEGMENTS']):
+      cumNumber = 0
+      while cumNumber < len(segment):
+        num = min(len(segment) - cumNumber, maxCount)
+        indexes = range(cumNumber, cumNumber + num)
+        cumNumber += num
+        lats = [str(segment[i][gpx.LAT]) for i in indexes]
+        lons = [str(segment[i][gpx.LON]) for i in indexes]
+        request = 'http://api.geonames.org/srtm1?lats=' + ','.join(lats) + '&lngs=' + ','.join(lons) + '&username=gpxviewer'
+        self.requestedIndexes[request] = (track, s, indexes)
+        self.networkManager.get(QNetworkRequest(QUrl(request)))
+        self.ui.statusBar.showMessage(self.tr('Getting altitudes from SRTM data...'))
+
+  @pyqtSlot()
+  def resetTrackAltitudes(self):
+    TheDocument.trkmodel.resetAltitudes([i.row() for i in self.ui.trkView.selectionModel().selectedRows()])
+    self.ui.statusBar.showMessage(self.tr('Altitudes reset to original values.'), 2000)
 
   @pyqtSlot()
   def showGoogleMap(self):
-    lat = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data()
-    lon = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data()
+    lat = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data(gpx.ValueRole))
+    lon = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data(gpx.ValueRole))
     webbrowser.open('https://maps.google.com/maps?ll=' + lat + ',' + lon + '&t=h&q=' + lat + ',' + lon + '&z=15')
 
   @pyqtSlot()
   def showYandexMap(self):
-    lat = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data()
-    lon = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data()
+    lat = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data(gpx.ValueRole))
+    lon = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data(gpx.ValueRole))
     webbrowser.open('https://maps.yandex.ru?ll=' + lon + ',' + lat + '&spn=0.03,0.03&pt=' + lon + ',' + lat + '&l=sat')
 
   @pyqtSlot()
   def showZoomEarthMap(self):
-    lat = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data()
-    lon = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data()
+    lat = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data(gpx.ValueRole))
+    lon = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data(gpx.ValueRole))
     webbrowser.open('https://zoom.earth/#' + lat + ',' + lon + ',15z,map')
 
   @pyqtSlot()
   def showOpenCycleMap(self):
-    lat = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data()
-    lon = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data()
+    lat = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data(gpx.ValueRole))
+    lon = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data(gpx.ValueRole))
     webbrowser.open('https://openstreetmap.org/?mlat=' + lat + '&mlon=' + lon + '&zoom=15&layers=C')
 
   @pyqtSlot()
   def showOpenTopoMap(self):
-    lat = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data()
-    lon = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data()
+    lat = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data(gpx.ValueRole))
+    lon = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data(gpx.ValueRole))
     webbrowser.open('https://opentopomap.org/#marker=15/' + lat + '/' + lon)
 
   @pyqtSlot()
   def showTopoMap(self):
-    lat = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data()
-    lon = TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data()
+    lat = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LAT).data(gpx.ValueRole))
+    lon = str(TheDocument.wptmodel.index(self.ui.wptView.currentIndex().data(gpx.IDRole), gpx.LON).data(gpx.ValueRole))
     webbrowser.open('http://loadmap.net/ru?q=' + lat + ' ' + lon + '&z=13&s=0')
 
   @pyqtSlot()
   def pointStyle(self):
     dlg = PointConfigDialog(self, self.ui.wptView.currentIndex().data(gpx.IDRole),
-                            [i.data(gpx.IDRole) for i in self.ui.wptView.selectedIndexes() if i.column() == gpx.NAME])
+                            [i.data(gpx.IDRole) for i in self.ui.wptView.selectionModel().selectedRows()])
     dlg.exec_()
 
   @pyqtSlot()
@@ -520,24 +629,37 @@ class GpxMainWindow(QMainWindow):
     self.reset()
 
   @pyqtSlot()
-  def onFileLoadGPXFile(self):
-    filenames = QFileDialog.getOpenFileNames(self, self.tr('Open GPX file'), TheConfig['MainWindow']['LoadGPXDirectory'],
-                                             self.tr('GPX XML (*.gpx *.GPX);;All files (*)'))[0]
+  def onFileLoadFile(self):
+    filenames = QFileDialog.getOpenFileNames(self, self.tr('Open file'), TheConfig['MainWindow']['LoadGPXDirectory'],
+                                             self.tr('All supported files (*.gpx *.GPX *.kml *.KML);;'
+                                                     'GPX XML (*.gpx *.GPX);;'
+                                                     'Keyhole Markup Language (*.kml *.KML);;'
+                                                     'All files (*)'))[0]
     self.openGPXFiles(filenames)
 
   @pyqtSlot()
-  def onFileSaveGPXFileAs(self):
+  def onFileSaveFileAs(self):
     if TheDocument.wptmodel.rowCount() == len(TheDocument.wptmodel.getSkippedPoints()) and \
        TheDocument.trkmodel.rowCount() == len(TheDocument.trkmodel.getSkippedTracks()):
-      QMessageBox.warning(self, self.tr('Save error'), self.tr('The GPX file will be empty.'))
+      QMessageBox.warning(self, self.tr('Save error'), self.tr('The file will be empty.'))
       return
 
-    filename = QFileDialog.getSaveFileName(self, self.tr('Save GPX file as'), TheConfig['MainWindow']['LoadGPXDirectory'],
-                                           self.tr('GPX XML (*.gpx *.GPX);;All files (*)'))[0]
+    filename, filter = QFileDialog.getSaveFileName(self, self.tr('Save file as'), TheConfig['MainWindow']['LoadGPXDirectory'],
+                                                   self.tr('GPX XML (*.gpx *.GPX);;'
+                                                           'Keyhole Markup Language (*.kml *.KML);;'
+                                                           'All files (*)'),
+                                                   TheConfig['MainWindow']['GPSFileExtension'])
     if filename != '':
-      TheConfig['MainWindow']['LoadGPXDirectory'] = QFileInfo(filename).path()
-      TheDocument.gpxparser.writeToFile(filename)
-      self.ui.statusBar.showMessage(filename + self.tr(' written.'), 2000)
+      try:
+        if filter.find('kml') != -1 or filename.lower().endswith('.kml'):
+          TheDocument.gpxparser.writeKMLFile(filename)
+        else:
+          TheDocument.gpxparser.writeGPXFile(filename)
+        TheConfig['MainWindow']['LoadGPXDirectory'] = QFileInfo(filename).path()
+        TheConfig['MainWindow']['GPSFileExtension'] = filter
+        self.ui.statusBar.showMessage(filename + self.tr(' written.'), 2000)
+      except OSError as e:
+        QMessageBox.warning(self, self.tr('Save error'), e.args[0])
 
   @pyqtSlot()
   def onFileOpen(self):
