@@ -14,8 +14,8 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import xml.etree.ElementTree as ET
 from datetime import datetime, timedelta
+from lxml import etree
 from math import acos, cos, modf, pi, sin, sqrt
 from PyQt5.QtCore import Qt, QAbstractTableModel, QFileInfo, QModelIndex, QObject, QPointF, QSortFilterProxyModel, pyqtSignal
 from PyQt5.QtGui import QColor, QFont, QGuiApplication, QPainter, QPainterPath, QPen, QPixmap, QPolygonF
@@ -23,10 +23,12 @@ from gpxviewer.configstore import TheConfig
 
 
 WPTFIELDS = NAME, LAT, LON, ALT, DIST, TIME, TIME_DELTA, TIME_DAYS, DIST_DELTA, ALT_DELTA, SPEED, ALT_SPEED, SLOPE = range(13)
-TRKFIELDS = TRKNAME, TRKSEGS, TRKPTS, TRKLEN, TRKTIME, TRKDUR = range(6)
-ValueRole, IDRole, IncludeRole, MarkerRole, CaptionRole, SplitLineRole, NeglectRole, MarkerStyleRole, CaptionStyleRole, SplitLineStyleRole = range(Qt.UserRole, Qt.UserRole + 10)
+TRKFIELDS = TRKNAME, TRKSEGS, TRKPTS, TRKLEN, TRKALTGAIN, TRKALTDROP, TRKTIME, TRKDUR = range(8)
+ValueRole, IDRole, IncludeRole, MarkerRole, CaptionRole, SplitLineRole, NeglectRole,\
+  MarkerStyleRole, CaptionStyleRole, SplitLineStyleRole = range(Qt.UserRole, Qt.UserRole + 10)
 MARKER_COLOR, MARKER_STYLE, MARKER_SIZE, CAPTION_POSX, CAPTION_POSY, CAPTION_ROTATION, CAPTION_SIZE, LINE_COLOR, LINE_STYLE, LINE_WIDTH = \
-  ('MarkerColor', 'MarkerStyle', 'MarkerSize', 'CaptionPositionX', 'CaptionPositionY', 'CaptionRotation', 'CaptionSize', 'SplitLineColor', 'SplitLineStyle', 'SplitLineWidth')
+  ('MarkerColor', 'MarkerStyle', 'MarkerSize', 'CaptionPositionX', 'CaptionPositionY',
+   'CaptionRotation', 'CaptionSize', 'SplitLineColor', 'SplitLineStyle', 'SplitLineWidth')
 
 
 class GpxWarning(Exception):
@@ -44,7 +46,8 @@ class WptModel(QAbstractTableModel):
 
     self.fields = [self.tr('Name'), self.tr('Latitude'), self.tr('Longitude'), self.tr('Altitude (m)'),
                    self.tr('Distance (km)'), self.tr('Time'), self.tr('Time difference'), self.tr('Time in days'),
-                   self.tr('Distance difference (km)'), self.tr('Altitude difference (m)'), self.tr('Speed (km/h)'), self.tr('Climbing speed (m/h)'), self.tr('Slope (m/km)')]
+                   self.tr('Distance difference (km)'), self.tr('Altitude difference (m)'),
+                   self.tr('Speed (km/h)'), self.tr('Climbing speed (m/h)'), self.tr('Slope (m/km)')]
     self.resetModel()
     self.pix = QPixmap(16, 16)
     self.pix.fill(Qt.transparent)
@@ -86,6 +89,8 @@ class WptModel(QAbstractTableModel):
           sec, min = modf(min)
           sec = round(sec * 60, 2)
           return str(int(deg)) + '°' + str(int(min)) + '\'' + str(sec) + '"'
+      elif index.column() == ALT and index.row() in self.changedAltitudes:
+        return str(self.changedAltitudes[index.row()])
       elif index.column() in {ALT, ALT_DELTA, ALT_SPEED, SLOPE} and self.waypoints[index.row()][index.column()] != '':
         return str(round(self.waypoints[index.row()][index.column()]))
       elif index.column() in {DIST, TIME_DAYS, DIST_DELTA, SPEED} and self.waypoints[index.row()][index.column()] != '':
@@ -145,7 +150,7 @@ class WptModel(QAbstractTableModel):
     return None
 
   def setData(self, index, value, role):
-    if index.isValid() and role == Qt.EditRole and value != self.waypoints[index.row()][NAME]:
+    if index.isValid() and role == Qt.EditRole and index.column() == NAME and value != self.waypoints[index.row()][NAME]:
       self.changedNames[index.row()] = value
       self.dataChanged.emit(index, index)
       self.wptDataChanged.emit()
@@ -210,6 +215,7 @@ class WptModel(QAbstractTableModel):
     self.neglectStates = []
     self.pointStyles = []
     self.changedNames = {}
+    self.changedAltitudes = {}
     self.endResetModel()
 
   def setIncludeStates(self, IDs, state, update=True):
@@ -265,7 +271,26 @@ class WptModel(QAbstractTableModel):
       if i in self.changedNames:
         del self.changedNames[i]
         self.dataChanged.emit(self.index(i, NAME), self.index(i, NAME))
-        self.wptDataChanged.emit()
+    self.wptDataChanged.emit()
+
+  def setAltitudes(self, IDs, alts):
+    for i, a in zip(IDs, alts):
+      self.changedAltitudes[i] = a
+      self.parent().updateMinMaxAltitudes(a)
+      self.dataChanged.emit(self.index(i, ALT), self.index(i, SLOPE))
+
+    self.parent().updateDetailedData()
+    self.wptDataChanged.emit()
+
+  def resetAltitudes(self, IDs):
+    for i in IDs:
+      if i in self.changedAltitudes:
+        del self.changedAltitudes[i]
+        self.parent().updateMinMaxAltitudes(self.waypoints[i][ALT])
+        self.dataChanged.emit(self.index(i, ALT), self.index(i, SLOPE))
+
+    self.parent().updateDetailedData()
+    self.wptDataChanged.emit()
 
   wptDataChanged = pyqtSignal()
 
@@ -273,7 +298,8 @@ class WptModel(QAbstractTableModel):
 class TrkModel(QAbstractTableModel):
   def __init__(self, parent=None):
     super(TrkModel, self).__init__(parent)
-    self.fields = [self.tr('Name'), self.tr('Segments'), self.tr('Points'), self.tr('Length'), self.tr('Time'), self.tr('Duration')]
+    self.fields = [self.tr('Name'), self.tr('Segments'), self.tr('Points'), self.tr('Length (km)'),
+                   self.tr('Altitude gain (m)'), self.tr('Altitude drop (m)'), self.tr('Start time'), self.tr('Duration')]
     self.resetModel()
 
     # Define colors that should look well for any color theme
@@ -296,6 +322,10 @@ class TrkModel(QAbstractTableModel):
     if role == Qt.DisplayRole:
       if index.column() == TRKTIME and self.tracks[index.row()][index.column()] != '':
         return str(self.tracks[index.row()][index.column()] + timedelta(minutes=TheConfig.getValue('ProfileStyle', 'TimeZoneOffset')))
+      elif index.column() in {TRKALTGAIN, TRKALTDROP}:
+        return str(round(self.tracks[index.row()][index.column()]))
+      elif index.column() == TRKLEN:
+        return str(round(self.tracks[index.row()][index.column()], 3))
       else:
         return str(self.tracks[index.row()][index.column()])
     elif role == IncludeRole:
@@ -304,6 +334,12 @@ class TrkModel(QAbstractTableModel):
       return QGuiApplication.palette().base().color() if self.includeStates[index.row()] else self.SkipColor
     return None
 
+  def getPointData(self, track, segment, index, column):
+    if column == ALT and track in self.changedAltitudes:
+      return self.changedAltitudes[track][(segment, index)]
+    else:
+      return self.tracks[track]['SEGMENTS'][segment][index][column]
+
   def headerData(self, section, orientation, role):
     if role == Qt.DisplayRole:
       return self.fields[section] if orientation == Qt.Horizontal else section + 1
@@ -318,10 +354,14 @@ class TrkModel(QAbstractTableModel):
   def getSkippedTracks(self):
     return [i for i, s in enumerate(self.includeStates) if not s]
 
+  def getChangedAltitudes(self):
+    return [v[(s, i)] for t, v in self.changedAltitudes.items() for s, seg in enumerate(self.tracks[t]['SEGMENTS']) for i in range(len(seg))]
+
   def resetModel(self):
     self.beginResetModel()
     self.tracks = []
     self.includeStates = []
+    self.changedAltitudes = {}
     self.endResetModel()
 
   def setIncludeStates(self, IDs, state, update=True):
@@ -331,6 +371,58 @@ class TrkModel(QAbstractTableModel):
     if update:
       self.parent().updatePoints()
 
+  def setTracksAltitudes(self, tracks, alts):
+    n = 0
+    for tnum, track in enumerate(tracks):
+      self.changedAltitudes[track] = {}
+      for s, seg in enumerate(self.tracks[track]['SEGMENTS']):
+        for i in range(len(seg)):
+          self.changedAltitudes[track][(s, i)] = alts[n]
+          n += 1
+
+      self.updateAltitudeGainDrop(tnum)
+
+  def setPointsAltitudes(self, track, segment, IDs, alts):
+    if track not in self.changedAltitudes:
+      self.changedAltitudes[track] = {}
+    for i, a in zip(IDs, alts):
+      self.changedAltitudes[track][(segment, i)] = a
+      self.parent().updateMinMaxAltitudes(a)
+
+    # Update only if all data is downloaded
+    if len(self.changedAltitudes[track]) == self.tracks[track][TRKPTS]:
+      self.updateAltitudeGainDrop(track)
+      self.dataChanged.emit(self.index(track, TRKALTGAIN), self.index(track, TRKALTDROP))
+    self.trkDataChanged.emit()
+
+  def resetAltitudes(self, tracks):
+    for t in tracks:
+      if t in self.changedAltitudes:
+        del self.changedAltitudes[t]
+      self.updateAltitudeGainDrop(t)
+      self.dataChanged.emit(self.index(t, TRKALTGAIN), self.index(t, TRKALTDROP))
+
+    self.trkDataChanged.emit()
+
+  def updateAltitudeGainDrop(self, track):
+    gain = 0
+    drop = 0
+    for snum, seg in enumerate(self.tracks[track]['SEGMENTS']):
+      prev_alt = None
+      for p in range(len(seg)):
+        if prev_alt is not None:
+          delta = self.getPointData(track, snum, p, ALT) - prev_alt
+          if delta >= 0:
+            gain += delta
+          else:
+            drop += delta
+        prev_alt = self.getPointData(track, snum, p, ALT)
+
+    self.tracks[track][TRKALTGAIN] = gain
+    self.tracks[track][TRKALTDROP] = drop
+
+  trkDataChanged = pyqtSignal()
+
 
 class GpxParser(QObject):
   def __init__(self, parent=None):
@@ -344,134 +436,245 @@ class GpxParser(QObject):
     self.trkmodel.resetModel()
     self.points = []
     self.minalt = 10000
-    self.maxalt = 0
+    self.maxalt = -10000
 
   def parse(self, filename):
     try:
-      doc = ET.parse(filename)
-    except (ET.ParseError, IsADirectoryError, FileNotFoundError):
-      raise GpxWarning(QFileInfo(filename).absoluteFilePath() + self.tr(' is an invalid GPX file.'))
-    ns = {'ns': doc.getroot().tag.split('}')[0][1:]}
+      doc = etree.parse(filename)
+    except (etree.ParseError, IsADirectoryError, FileNotFoundError):
+      raise GpxWarning(QFileInfo(filename).absoluteFilePath() + self.tr(' is an invalid file.'))
 
+    self.ns = doc.getroot().nsmap
+    self.ns['ns'] = self.ns[None]
+
+    if filename.lower().endswith('.kml'):
+      self.parseKMLDocument(doc.getroot().find('{%(ns)s}Document' % self.ns))
+    else:
+      self.parseGPXDocument(doc.getroot())
+
+    self.updateMinMaxAltitudes()
+
+  def parseGPXDocument(self, root):
     tag = 'name'
     if TheConfig.getValue('ProfileStyle', 'ReadNameFromTag') == 1:  # Comment
       tag = 'cmt'
     elif TheConfig.getValue('ProfileStyle', 'ReadNameFromTag') == 2:  # Description
       tag = 'desc'
 
-    for p in doc.iterfind('.//{%(ns)s}wpt' % ns):
+    for p in root.iterfind('{%(ns)s}wpt' % self.ns):
       try:
-        point = {}
-        name = p.findtext(('{%(ns)s}' + tag) % ns)
-        point[NAME] = name.strip() if name is not None else ''
-        point[LAT] = float(p.get('lat'))
-        point[LON] = float(p.get('lon'))
-        ele = p.findtext('{%(ns)s}ele' % ns)
-        point[ALT] = float(ele) if ele is not None else 0
-        self.minalt = min(self.minalt, point[ALT])
-        self.maxalt = max(self.maxalt, point[ALT])
-        time = p.findtext('{%(ns)s}time' % ns)
-        if time is not None:
-          point[TIME] = datetime.strptime(time[:19].strip(), '%Y-%m-%dT%H:%M:%S')
-        else:
-          point[TIME] = ''
-        point[DIST] = ''
-        point[TIME_DELTA] = ''
-        point[TIME_DAYS] = ''
-        point[DIST_DELTA] = ''
-        point[ALT_DELTA] = ''
-        point[SPEED] = ''
-        point[ALT_SPEED] = ''
-        point[SLOPE] = ''
+        name = p.findtext(('{%(ns)s}' + tag) % self.ns)
+        ele = p.findtext('{%(ns)s}ele' % self.ns)
+        time = p.findtext('{%(ns)s}time' % self.ns)
+        point = {
+          NAME: name.strip() if name is not None else '',
+          LAT: float(p.get('lat')),
+          LON: float(p.get('lon')),
+          ALT: float(ele) if ele is not None else 0,
+          TIME: datetime.fromisoformat(time[0:-1]) if time is not None else '',
+          # Additional fields
+          'CMT': p.findtext('{%(ns)s}cmt' % self.ns),
+          'DESC': p.findtext('{%(ns)s}desc' % self.ns),
+          'SYM': p.findtext('{%(ns)s}sym' % self.ns),
+        }
 
-        wptid = self.wptmodel.rowCount()
-        # Sort points by time
-        if TheConfig.getValue('ProfileStyle', 'SortByTime') and point[TIME] != '':
-          while wptid > 0 and self.wptmodel.waypoints[wptid - 1][TIME] != '' and point[TIME] < self.wptmodel.waypoints[wptid - 1][TIME]:
-            wptid -= 1
-          for i in range(wptid, self.wptmodel.rowCount()):
-            self.wptmodel.waypoints[i]['ID'] += 1
-        point['ID'] = wptid
-
-        # Additional fields
-        point['CMT'] = p.findtext('{%(ns)s}cmt' % ns)
-        point['DESC'] = p.findtext('{%(ns)s}desc' % ns)
-        point['SYM'] = p.findtext('{%(ns)s}sym' % ns)
-
-        self.wptmodel.beginInsertRows(QModelIndex(), wptid, wptid)
-        self.wptmodel.waypoints.insert(wptid, point)
-        self.wptmodel.includeStates.insert(wptid, True)
-        self.wptmodel.markerStates.insert(wptid, False)
-        self.wptmodel.captionStates.insert(wptid, False)
-        self.wptmodel.splitStates.insert(wptid, False)
-        self.wptmodel.neglectStates.insert(wptid, False)
-        self.wptmodel.pointStyles.insert(wptid, {})
-        self.wptmodel.endInsertRows()
+        self.addPointToModel(point)
 
       except (TypeError, ValueError):
         self.warningSent.emit(self.tr('File read error'),
-                              self.tr('Waypoint ') + (point[NAME] + ' ' if point[NAME] != '' else '') + self.tr('is invalid and will be skipped.', 'Waypoint'))
+                              self.tr('Waypoint ') + (name + ' ' if name is not None else '') +
+                              self.tr('is invalid and will be skipped.', 'Waypoint'))
 
-    trkid = self.trkmodel.rowCount()
-    for t in doc.iterfind('.//{%(ns)s}trk' % ns):
+    for t in root.iterfind('{%(ns)s}trk' % self.ns):
       try:
-        track = {}
-        name = t.findtext('{%(ns)s}name' % ns)
-        track[TRKNAME] = name.strip() if name is not None else ''
+        name = t.findtext('{%(ns)s}name' % self.ns)
+        track = {TRKNAME: name.strip() if name is not None else ''}
 
-        track['SEGMENTS'] = []
         pts = 0
-        tot_dist = 0.0
-        for s in t.iterfind('.//{%(ns)s}trkseg' % ns):
-          segment = []
+        dist = 0.0
+        track['SEGMENTS'] = []
+        for s in t.iterfind('{%(ns)s}trkseg' % self.ns):
           prev_lat = None
           prev_lon = None
-          dist = 0.0
-          for p in s.iterfind('.//{%(ns)s}trkpt' % ns):
-            point = {}
-            point[LAT] = float(p.get('lat'))
-            point[LON] = float(p.get('lon'))
-            ele = p.findtext('{%(ns)s}ele' % ns)
-            point[ALT] = float(ele) if ele is not None else 0
+          segment = []
+          for p in s.iterfind('{%(ns)s}trkpt' % self.ns):
+            ele = p.findtext('{%(ns)s}ele' % self.ns)
+            time = p.findtext('{%(ns)s}time' % self.ns)
+            point = {
+              LAT: float(p.get('lat')),
+              LON: float(p.get('lon')),
+              ALT: float(ele) if ele is not None else 0,
+              TIME: datetime.fromisoformat(time[0:-1]) if time is not None else '',
+            }
             self.minalt = min(self.minalt, point[ALT])
             self.maxalt = max(self.maxalt, point[ALT])
-            time = p.findtext('{%(ns)s}time' % ns)
-            if time is not None:
-              point[TIME] = datetime.strptime(time[:19].strip(), '%Y-%m-%dT%H:%M:%S')
-            else:
-              point[TIME] = ''
 
             if prev_lat is not None and prev_lon is not None:
               dist += _distance(point[LAT], point[LON], prev_lat, prev_lon)
-            prev_lat = point[LAT]
-            prev_lon = point[LON]
+            prev_lat, prev_lon = point[LAT], point[LON]
 
             segment += [point]
             pts += 1
 
           track['SEGMENTS'] += [segment]
-          tot_dist += dist
 
-        track[TRKSEGS] = len(track['SEGMENTS'])
         track[TRKPTS] = pts
-        track['LENGTH'] = tot_dist
-        track[TRKLEN] = ''
-        track[TRKTIME] = track['SEGMENTS'][0][0][TIME]
-        track[TRKDUR] = track['SEGMENTS'][-1][-1][TIME] - track['SEGMENTS'][0][0][TIME] \
-                        if track['SEGMENTS'][0][0][TIME] != '' and track['SEGMENTS'][-1][-1][TIME] != '' else ''
-
-        self.trkmodel.beginInsertRows(QModelIndex(), trkid, trkid)
-        self.trkmodel.tracks += [track]
-        self.trkmodel.includeStates += [True]
-        self.trkmodel.endInsertRows()
-        trkid += 1
+        track['LENGTH'] = dist
+        self.addTrackToModel(track)
 
       except (TypeError, ValueError):
         self.warningSent.emit(self.tr('File read error'),
-                              self.tr('Track ') + (track[TRKNAME] + ' ' if track[TRKNAME] != '' else '') + self.tr('is invalid and will be skipped.', 'Track'))
+                              self.tr('Track ') + (name + ' ' if name is not None else '') +
+                              self.tr('is invalid and will be skipped.', 'Track'))
 
-    TheConfig['ProfileStyle']['MinimumAltitude'] = str(int(round(self.minalt, -3) - 500 if round(self.minalt, -3) > self.minalt else round(self.minalt, -3)))
-    TheConfig['ProfileStyle']['MaximumAltitude'] = str(int(round(self.maxalt, -3) + 500 if round(self.maxalt, -3) < self.maxalt else round(self.maxalt, -3)))
+  def parseKMLDocument(self, root):
+    for element in root.iterchildren():
+      if element.tag == '{%(ns)s}Folder' % self.ns:
+        self.parseKMLDocument(element)
+
+      elif element.tag == '{%(ns)s}Placemark' % self.ns:
+        if element.find('{%(ns)s}Point' % self.ns) is not None:  # waypoint
+          try:
+            name = element.findtext('{%(ns)s}name' % self.ns)
+            time = element.findtext('.//{%(ns)s}when' % self.ns)
+            point = {
+              NAME: name.strip() if name is not None else '',
+              TIME: datetime.fromisoformat(time[0:-1]) if time is not None else '',
+            }
+            point[LON], point[LAT], point[ALT] = eval(element.findtext('.//{%(ns)s}coordinates' % self.ns))
+
+            self.addPointToModel(point)
+
+          except (TypeError, ValueError):
+            self.warningSent.emit(self.tr('File read error'),
+                                  self.tr('Waypoint ') + (name + ' ' if name is not None else '') +
+                                  self.tr('is invalid and will be skipped.', 'Waypoint'))
+
+        elif element.find('{%(gx)s}Track' % self.ns) is not None:  # track
+          try:
+            name = element.findtext('{%(ns)s}name' % self.ns)
+            track = {TRKNAME: name.strip() if name is not None else ''}
+
+            coords = [el.text.split(' ') for el in element.iterfind('.//{%(gx)s}coord' % self.ns)]
+            times = [datetime.fromisoformat(el.text[0:-1]) for el in element.iterfind('.//{%(ns)s}when' % self.ns)]
+            if len(coords) != len(times):
+              times = [None] * len(coords)
+
+            prev_lat = None
+            prev_lon = None
+            dist = 0.0
+            segment = []
+            for c, t in zip(coords, times):
+              point = {
+                LON: float(c[0]),
+                LAT: float(c[1]),
+                ALT: float(c[2]),
+                TIME: t if t is not None else ''
+              }
+              self.minalt = min(self.minalt, point[ALT])
+              self.maxalt = max(self.maxalt, point[ALT])
+
+              if prev_lat is not None and prev_lon is not None:
+                dist += _distance(point[LAT], point[LON], prev_lat, prev_lon)
+              prev_lat, prev_lon = point[LAT], point[LON]
+
+              segment += [point]
+
+            track['SEGMENTS'] = [segment]
+            track[TRKPTS] = len(segment)
+            track['LENGTH'] = dist
+            self.addTrackToModel(track)
+
+          except (TypeError, ValueError):
+            self.warningSent.emit(self.tr('File read error'),
+                                  self.tr('Track ') + (name + ' ' if name is not None else '') +
+                                  self.tr('is invalid and will be skipped.', 'Track'))
+
+        elif element.find('{%(ns)s}LineString' % self.ns) is not None:  # line
+          try:
+            name = element.findtext('{%(ns)s}name' % self.ns)
+            track = {TRKNAME: name.strip() if name is not None else ''}
+
+            coords = element.findtext('.//{%(ns)s}coordinates' % self.ns)
+            coords = coords.strip().split(' ') if coords is not None else []
+            times = [None] * len(coords)
+
+            prev_lat = None
+            prev_lon = None
+            dist = 0.0
+            segment = []
+            for c, t in zip(coords, times):
+              point = {TIME: ''}
+              point[LON], point[LAT], point[ALT] = [float(n) for n in eval(c)]
+              self.minalt = min(self.minalt, point[ALT])
+              self.maxalt = max(self.maxalt, point[ALT])
+
+              if prev_lat is not None and prev_lon is not None:
+                dist += _distance(point[LAT], point[LON], prev_lat, prev_lon)
+              prev_lat, prev_lon = point[LAT], point[LON]
+
+              segment += [point]
+
+            track['SEGMENTS'] = [segment]
+            track[TRKPTS] = len(segment)
+            track['LENGTH'] = dist
+            self.addTrackToModel(track)
+
+          except (TypeError, ValueError):
+            self.warningSent.emit(self.tr('File read error'),
+                                  self.tr('Track ') + (name + ' ' if name is not None else '') +
+                                  self.tr('is invalid and will be skipped.', 'Track'))
+
+  def addPointToModel(self, point):
+    point[DIST] = ''
+    point[TIME_DELTA] = ''
+    point[TIME_DAYS] = ''
+    point[DIST_DELTA] = ''
+    point[ALT_DELTA] = ''
+    point[SPEED] = ''
+    point[ALT_SPEED] = ''
+    point[SLOPE] = ''
+
+    wptid = self.wptmodel.rowCount()
+    # Sort points by time
+    if TheConfig.getValue('ProfileStyle', 'SortByTime') and point[TIME] != '':
+      while wptid > 0 and self.wptmodel.waypoints[wptid - 1][TIME] != '' and point[TIME] < self.wptmodel.waypoints[wptid - 1][TIME]:
+        wptid -= 1
+      for i in range(wptid, self.wptmodel.rowCount()):
+        self.wptmodel.waypoints[i]['ID'] += 1
+    point['ID'] = wptid
+
+    self.wptmodel.beginInsertRows(QModelIndex(), wptid, wptid)
+    self.wptmodel.waypoints.insert(wptid, point)
+    self.wptmodel.includeStates.insert(wptid, True)
+    self.wptmodel.markerStates.insert(wptid, False)
+    self.wptmodel.captionStates.insert(wptid, False)
+    self.wptmodel.splitStates.insert(wptid, False)
+    self.wptmodel.neglectStates.insert(wptid, False)
+    self.wptmodel.pointStyles.insert(wptid, {})
+    self.wptmodel.endInsertRows()
+
+    self.minalt = min(self.minalt, point[ALT])
+    self.maxalt = max(self.maxalt, point[ALT])
+
+  def addTrackToModel(self, track):
+    track[TRKSEGS] = len(track['SEGMENTS'])
+    track[TRKLEN] = ''
+    track[TRKALTGAIN] = ''
+    track[TRKALTDROP] = ''
+    if all([p[TIME] != '' for s in track['SEGMENTS'] for p in s]):
+      track[TRKTIME] = track['SEGMENTS'][0][0][TIME]
+      track[TRKDUR] = track['SEGMENTS'][-1][-1][TIME] - track['SEGMENTS'][0][0][TIME]
+    else:
+      track[TRKTIME] = ''
+      track[TRKDUR] = ''
+
+    trkid = self.trkmodel.rowCount()
+    self.trkmodel.beginInsertRows(QModelIndex(), trkid, trkid)
+    self.trkmodel.tracks += [track]
+    self.trkmodel.includeStates += [True]
+    self.trkmodel.updateAltitudeGainDrop(trkid)
+    self.trkmodel.endInsertRows()
 
   def updatePoints(self):
     self.points = []
@@ -490,7 +693,10 @@ class GpxParser(QObject):
       i = 0
       while i < self.wptmodel.rowCount():
         if self.wptmodel.includeStates[i]:
-          if ind < len(self.points) and self.wptmodel.waypoints[i][TIME] < self.trkmodel.tracks[self.points[ind][0]]['SEGMENTS'][self.points[ind][1]][self.points[ind][2]][TIME] or ind == len(self.points):
+          if ind < len(self.points) and \
+             self.wptmodel.waypoints[i][TIME] < \
+             self.trkmodel.getPointData(self.points[ind][0], self.points[ind][1], self.points[ind][2], TIME) or \
+             ind == len(self.points):
             self.points.insert(ind, i)
             i += 1
           ind += 1
@@ -520,8 +726,8 @@ class GpxParser(QObject):
         prev_lat = lat
         prev_lon = lon
       else:
-        lat = self.trkmodel.tracks[p[0]]['SEGMENTS'][p[1]][p[2]][LAT]
-        lon = self.trkmodel.tracks[p[0]]['SEGMENTS'][p[1]][p[2]][LON]
+        lat = self.trkmodel.getPointData(p[0], p[1], p[2], LAT)
+        lon = self.trkmodel.getPointData(p[0], p[1], p[2], LON)
         if prev_lat is not None and prev_lon is not None:
           dist += _distance(lat, lon, prev_lat, prev_lon) * dist_coeff
         self.trkmodel.tracks[p[0]]['SEGMENTS'][p[1]][p[2]][DIST] = dist
@@ -532,7 +738,7 @@ class GpxParser(QObject):
       self.wptmodel.waypoints[i][DIST] = ''
 
     for t in self.trkmodel.tracks:
-      t[TRKLEN] = round(t['LENGTH'] * dist_coeff, 3)
+      t[TRKLEN] = t['LENGTH'] * dist_coeff
 
     self.updateDetailedData()
 
@@ -591,7 +797,7 @@ class GpxParser(QObject):
       if self.wptmodel.includeStates[j]:
         dt = (p[TIME_DAYS] - prev[TIME_DAYS]) * 24.0 if p[TIME_DAYS] != '' and prev[TIME_DAYS] != '' and prev[TIME_DAYS] != 0 else 0
         p[DIST_DELTA] = p[DIST] - prev[DIST]
-        p[ALT_DELTA] = p[ALT] - prev[ALT]
+        p[ALT_DELTA] = int(self.wptmodel.index(p['ID'], ALT).data()) - int(self.wptmodel.index(prev['ID'], ALT).data())
         p[SPEED] = p[DIST_DELTA] / dt if dt != 0 else 0.0
         p[ALT_SPEED] = p[ALT_DELTA] / dt if dt != 0 else 0.0
         p[SLOPE] = p[ALT_DELTA] / p[DIST_DELTA] if p[DIST_DELTA] != 0 else 0.0
@@ -603,106 +809,140 @@ class GpxParser(QObject):
         p[ALT_SPEED] = ''
         p[SLOPE] = ''
 
-  def writeToFile(self, filename):
-    root = ET.Element('gpx', attrib={'version': '1.1',
-                                     'creator': 'GPXViewer - https://bitbucket.org/salsergey/gpxviewer',
-                                     'xmlns': 'http://www.topografix.com/GPX/1/1'})
-    metadata = ET.Element('metadata')
-    root.append(metadata)
+  def updateMinMaxAltitudes(self, alt=None):
+    if alt is not None:
+      self.minalt = min(TheConfig.getValue('ProfileStyle', 'MinimumAltitude'), alt)
+      self.maxalt = max(TheConfig.getValue('ProfileStyle', 'MaximumAltitude'), alt)
+
+    TheConfig['ProfileStyle']['MinimumAltitude'] = str(int(round(self.minalt, -3) - 500 if round(self.minalt, -3) > self.minalt
+                                                           else round(self.minalt, -3)))
+    TheConfig['ProfileStyle']['MaximumAltitude'] = str(int(round(self.maxalt, -3) + 500 if round(self.maxalt, -3) < self.maxalt
+                                                           else round(self.maxalt, -3)))
+
+  def writeGPXFile(self, filename):
+    ns = {None: 'http://www.topografix.com/GPX/1/1'}
+    root = etree.Element('gpx', nsmap=ns, version='1.1', creator='GPX Viewer - https://osdn.net/projects/gpxviewer')
+    metadata = etree.SubElement(root, 'metadata')
 
     minlat = 90
     minlon = 180
     maxlat = 0
     maxlon = -180
-    for p, s in zip(self.wptmodel.waypoints, self.wptmodel.includeStates):
-      if s:
-        element = ET.Element('wpt', attrib={'lat': str(p[LAT]), 'lon': str(p[LON])})
-        root.append(element)
-        if p[LAT] < minlat:
-          minlat = p[LAT]
-        if p[LON] < minlon:
-          minlon = p[LON]
-        if p[LAT] > maxlat:
-          maxlat = p[LAT]
-        if p[LON] > maxlon:
-          maxlon = p[LON]
-        el = ET.Element('ele')
-        el.text = str(p[ALT])
-        element.append(el)
-        if p[TIME] != '':
-          el = ET.Element('time')
-          el.text = p[TIME].strftime('%Y-%m-%dT%H:%M:%SZ')
-          element.append(el)
-        el = ET.Element('name')
-        if p['ID'] in self.wptmodel.changedNames:
-          el.text = self.wptmodel.changedNames[p['ID']]
-        else:
-          el.text = p[NAME]
-        element.append(el)
-        if p['CMT'] is not None:
-          el = ET.Element('cmt')
-          el.text = p['CMT']
-          element.append(el)
-        if p['DESC'] is not None:
-          el = ET.Element('desc')
-          el.text = p['DESC']
-          element.append(el)
-        if p['SYM'] is not None:
-          el = ET.Element('sym')
-          el.text = p['SYM']
-          element.append(el)
 
-    for track, s in zip(self.trkmodel.tracks, self.trkmodel.includeStates):
-      if s:
-        elTrk = ET.Element('trk')
-        root.append(elTrk)
-        el = ET.Element('name')
-        el.text = track[TRKNAME]
-        elTrk.append(el)
-        for seg in track['SEGMENTS']:
-          elSeg = ET.Element('trkseg')
-          elTrk.append(elSeg)
-          for p in seg:
-            elP = ET.Element('trkpt', attrib={'lat': str(p[LAT]), 'lon': str(p[LON])})
-            elSeg.append(elP)
-            if p[LAT] < minlat:
-              minlat = p[LAT]
-            if p[LON] < minlon:
-              minlon = p[LON]
-            if p[LAT] > maxlat:
-              maxlat = p[LAT]
-            if p[LON] > maxlon:
-              maxlon = p[LON]
-            el = ET.Element('ele')
-            el.text = str(p[ALT])
-            elP.append(el)
-            el = ET.Element('time')
-            el.text = p[TIME].strftime('%Y-%m-%dT%H:%M:%SZ')
-            elP.append(el)
+    for point, state in zip(self.wptmodel.waypoints, self.wptmodel.includeStates):
+      if state:
+        element = etree.SubElement(root, 'wpt', lat=str(point[LAT]), lon=str(point[LON]))
+        etree.SubElement(element, 'ele').text = self.wptmodel.index(point['ID'], ALT).data()
+        minlat = min(minlat, point[LAT])
+        minlon = min(minlon, point[LON])
+        maxlat = max(maxlat, point[LAT])
+        maxlon = max(maxlon, point[LON])
+        if point[TIME] != '':
+          etree.SubElement(element, 'time').text = point[TIME].isoformat() + 'Z'
+        etree.SubElement(element, 'name').text = self.wptmodel.index(point['ID'], NAME).data()
 
-    el = ET.Element('time')
-    el.text = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
-    metadata.append(el)
-    el = ET.Element('bounds', attrib={'minlat': str(minlat), 'minlon': str(minlon), 'maxlat': str(maxlat), 'maxlon': str(maxlon)})
-    metadata.append(el)
+        if 'CMT' in point and point['CMT'] is not None:
+          etree.SubElement(element, 'cmt').text = point['CMT']
+        if 'DESC' in point and point['DESC'] is not None:
+          etree.SubElement(element, 'desc').text = point['DESC']
+        if 'SYM' in point and point['SYM'] is not None:
+          etree.SubElement(element, 'sym').text = point['SYM']
+
+    for tnum, track, state in zip(range(len(self.trkmodel.tracks)), self.trkmodel.tracks, self.trkmodel.includeStates):
+      if state:
+        elTrk = etree.SubElement(root, 'trk')
+        etree.SubElement(elTrk, 'name').text = track[TRKNAME]
+        for snum, seg in enumerate(track['SEGMENTS']):
+          elSeg = etree.SubElement(elTrk, 'trkseg')
+          for i, point in enumerate(seg):
+            elP = etree.SubElement(elSeg, 'trkpt', lat=str(point[LAT]), lon=str(point[LON]))
+            etree.SubElement(elP, 'ele').text = str(self.trkmodel.getPointData(tnum, snum, i, ALT))
+            minlat = min(minlat, point[LAT])
+            minlon = min(minlon, point[LON])
+            maxlat = max(maxlat, point[LAT])
+            maxlon = max(maxlon, point[LON])
+            if point[TIME] != '':
+              etree.SubElement(elP, 'time').text = point[TIME].isoformat() + 'Z'
+
+    etree.SubElement(metadata, 'time').text = datetime.utcnow().isoformat() + 'Z'
+    etree.SubElement(metadata, 'bounds', minlat=str(minlat), minlon=str(minlon), maxlat=str(maxlat), maxlon=str(maxlon))
 
-    outgpx = '<?xml version="1.0" encoding="UTF-8"?>\n' + ET.tostring(root, encoding='unicode') + '\n'
-    outgpx = outgpx.replace('<metadata', '\n  <metadata')
-    outgpx = outgpx.replace('</metadata', '\n  </metadata')
-    outgpx = outgpx.replace('<wpt', '\n  <wpt')
-    outgpx = outgpx.replace('<trk', '\n  <trk')
-    outgpx = outgpx.replace('<bounds', '\n    <bounds')
-    outgpx = outgpx.replace('<ele', '\n    <ele')
-    outgpx = outgpx.replace('<time', '\n    <time')
-    outgpx = outgpx.replace('<name', '\n    <name')
-    outgpx = outgpx.replace('<cmt', '\n    <cmt')
-    outgpx = outgpx.replace('<desc', '\n    <desc')
-    outgpx = outgpx.replace('<sym', '\n    <sym')
-    outgpx = outgpx.replace('</wpt', '\n  </wpt')
-    outgpx = outgpx.replace('</trk', '\n  </trk')
-    outgpx = outgpx.replace('</gpx', '\n</gpx')
     with open(filename, 'w', encoding='utf-8') as file:
-      file.write(outgpx)
+      file.write('<?xml version="1.0" encoding="UTF-8"?>\n' + etree.tostring(root, encoding='unicode', pretty_print=True))
+
+  def writeKMLFile(self, filename):
+    ns = {
+      None: 'http://www.opengis.net/kml/2.2',
+      'gx': 'http://www.google.com/kml/ext/2.2',
+      'kml': 'http://www.opengis.net/kml/2.2',
+      'atom': 'http://www.w3.org/2005/Atom'
+    }
+    root = etree.Element('kml', nsmap=ns, creator='GPX Viewer - https://osdn.net/projects/gpxviewer')
+    document = etree.SubElement(root, 'Document')
+    etree.SubElement(document, 'name').text = QFileInfo(filename).fileName()
+    etree.SubElement(document, 'snippet').text = 'Created ' + datetime.utcnow().ctime()
+
+    # Define simple styles for tracks
+    trkStyle = etree.SubElement(document, 'Style', id='track_n')
+    lineStyle = etree.SubElement(trkStyle, 'LineStyle')
+    etree.SubElement(lineStyle, 'color').text = '99ffac59'
+    etree.SubElement(lineStyle, 'width').text = '6'
+    trkStyle = etree.SubElement(document, 'Style', id='track_h')
+    lineStyle = etree.SubElement(trkStyle, 'LineStyle')
+    etree.SubElement(lineStyle, 'color').text = '99ffac59'
+    etree.SubElement(lineStyle, 'width').text = '8'
+    trkStyleMap = etree.SubElement(document, 'StyleMap', id='track')
+    pair = etree.SubElement(trkStyleMap, 'Pair')
+    etree.SubElement(pair, 'key').text = 'normal'
+    etree.SubElement(pair, 'styleUrl').text = '#track_n'
+    pair = etree.SubElement(trkStyleMap, 'Pair')
+    etree.SubElement(pair, 'key').text = 'highlight'
+    etree.SubElement(pair, 'styleUrl').text = '#track_h'
+
+    wptfolder = etree.SubElement(document, 'Folder')
+    etree.SubElement(wptfolder, 'name').text = 'Waypoints'
+    trkfolder = etree.SubElement(document, 'Folder')
+    etree.SubElement(trkfolder, 'name').text = 'Tracks'
+
+    for point, state in zip(self.wptmodel.waypoints, self.wptmodel.includeStates):
+      if state:
+        place = etree.SubElement(wptfolder, 'Placemark')
+        etree.SubElement(place, 'name').text = self.wptmodel.index(point['ID'], NAME).data()
+        if point[TIME] != '':
+          ts = etree.SubElement(place, 'TimeStamp')
+          etree.SubElement(ts, 'when').text = point[TIME].isoformat() + 'Z'
+        p = etree.SubElement(place, 'Point')
+        etree.SubElement(p, 'coordinates').text = str(point[LON]) + ',' + str(point[LAT]) + ',' + self.wptmodel.index(point['ID'], ALT).data()
+
+    for tnum, track, state in zip(range(len(self.trkmodel.tracks)), self.trkmodel.tracks, self.trkmodel.includeStates):
+      if state:
+        place = etree.SubElement(trkfolder, 'Placemark')
+        etree.SubElement(place, 'name').text = track[TRKNAME]
+        etree.SubElement(place, 'styleUrl').text = '#track'
+
+        if track[TRKTIME] != '':
+          tr = etree.SubElement(place, '{%(gx)s}Track' % ns)
+          times = []
+          coords = []
+          for snum, seg in enumerate(track['SEGMENTS']):
+            for i, point in enumerate(seg):
+              times += [point[TIME].isoformat() + 'Z']
+              coords += [str(point[LON]) + ' ' + str(point[LAT]) + ' ' + str(self.trkmodel.getPointData(tnum, snum, i, ALT))]
+          for t in times:
+            etree.SubElement(tr, 'when').text = t
+          for c in coords:
+            etree.SubElement(tr, '{%(gx)s}coord' % ns).text = c
+
+        else:
+          tr = etree.SubElement(place, 'LineString')
+          coords = []
+          for snum, seg in enumerate(track['SEGMENTS']):
+            for i, point in enumerate(seg):
+              coords += [str(point[LON]) + ',' + str(point[LAT]) + ',' + str(self.trkmodel.getPointData(tnum, snum, i, ALT))]
+          etree.SubElement(tr, 'coordinates').text = ' '.join(coords)
+
+    with open(filename, 'w', encoding='utf-8') as file:
+      file.write('<?xml version="1.0" encoding="UTF-8"?>\n' + etree.tostring(root, encoding='unicode', pretty_print=True))
 
   warningSent = pyqtSignal(str, str)
 
@@ -762,7 +1002,7 @@ def _distance(lat1, lon1, lat2, lon2):
   Radius = 6378.14  # The equatorial radius of the Earth
   # angle between two points
   angle = acos(min(1.0, sin(lat1*pi/180.0) * sin(lat2*pi/180.0) +
-               cos(lat1*pi/180.0) * cos(lat2*pi/180.0) * cos((lon1 - lon2)*pi/180.0)))
+                   cos(lat1*pi/180.0) * cos(lat2*pi/180.0) * cos((lon1 - lon2)*pi/180.0)))
   # local radius of the Earth
   r = Radius * (0.99832407 + 0.00167644 * cos(2.0*lat1) - 0.00000352 * cos(4.0*lat1))
   dist = r * angle
